# -*- coding: utf-8 -*-

"""Utilities for BioKEEN."""

import importlib
import logging
import os
import sys
from contextlib import redirect_stdout
from typing import Optional, Union

import numpy as np
import pkg_resources

from bio2bel import AbstractManager
from bio2bel.manager.bel_manager import BELManagerMixin
from pybel import from_json_path, from_web, to_json_path
from .constants import EMOJI, biokeen_config
from .convert import to_pykeen_df, to_pykeen_path, to_pykeen_summary_path

_SPECIAL_CASES = {
    'compath': 'compath_resources',
}

logger = logging.getLogger(__name__)


def install_bio2bel_module(name: str, connection: Optional[str] = None, rebuild: bool = False) -> Optional[str]:
    """Install Bio2BEL module.

    :param name: The name of the Bio2BEL module
    :param connection: The optional database connection
    :param rebuild: Should the cache not be used? Defaults to False.
    """
    module_name = _SPECIAL_CASES.get(name, f'bio2bel_{name}')

    pykeen_df_path = os.path.join(biokeen_config.data_directory, f'{name}.{biokeen_config.keen_tsv_ext}')
    pykeen_df_summary_path = os.path.join(biokeen_config.data_directory, f'{name}.keen.summary.json')
    json_path = os.path.join(biokeen_config.data_directory, f'{name}.bel.json')

    if os.path.exists(pykeen_df_path) and not rebuild:
        logger.info(f'{EMOJI} {module_name} has already been retrieved. See: {pykeen_df_path}')
        return pykeen_df_path

    if os.path.exists(json_path) and not rebuild:
        logger.info(f'{EMOJI} loaded {module_name} JSON: {json_path}')
        graph = from_json_path(json_path)
        df = to_pykeen_df(graph)
        to_pykeen_path(df, pykeen_df_path)
        to_pykeen_summary_path(df, pykeen_df_summary_path)
        return pykeen_df_path

    bio2bel_module = ensure_bio2bel_installation(module_name)
    logger.debug(f'{EMOJI} imported {module_name}')

    manager_cls = bio2bel_module.Manager

    if not issubclass(manager_cls, BELManagerMixin):
        version = pkg_resources.get_distribution(module_name).version
        logger.warning(f'{EMOJI} {module_name} v{version} does not produce BEL')
        sys.exit(1)

    manager = manager_cls(connection=connection)

    if issubclass(manager_cls, AbstractManager):
        if not manager.is_populated():
            logger.info(f'{EMOJI} populating {module_name}')
            manager.populate()
        else:
            logger.debug(f'{EMOJI} {module_name} has already been populated')

    logger.debug(f'{EMOJI} generating BEL for {module_name}')
    graph = manager.to_bel()

    logger.debug(f'Summary: {graph.number_of_nodes()} nodes / {graph.number_of_edges()} edges')
    to_json_path(graph, json_path, indent=2)

    logger.debug(f'{EMOJI} generating PyKEEN TSV for {module_name}')
    df = to_pykeen_df(graph)
    to_pykeen_summary_path(df, pykeen_df_summary_path)
    success = to_pykeen_path(df, pykeen_df_path)

    if success:
        logger.debug(f'{EMOJI} wrote PyKEEN TSV to {pykeen_df_path}')
        return pykeen_df_path

    logger.warning(f'{EMOJI} no statements generated')


def ensure_bio2bel_installation(package: str):
    """Import a package, or install it."""
    try:
        b_module = importlib.import_module(package)

    except ImportError:
        logger.info(f'{EMOJI} pip install {package}')
        # Install this package using pip
        # https://stackoverflow.com/questions/12332975/installing-python-module-within-code
        from pip._internal import main as pip_main

        with redirect_stdout(sys.stderr):
            pip_exit_code = pip_main(['install', '-q', package])  # -q means quiet

        if 0 != pip_exit_code:  # command failed
            logger.warning(f'{EMOJI} could not find {package} on PyPI. Try installing from GitHub with:')
            name = package.split("_")[-1]
            logger.warning(f'\n   pip install git+https://github.com/bio2bel/{name}.git\n')
            sys.exit(1)

        try:
            return importlib.import_module(package)
        except ImportError:
            logger.exception(f'{EMOJI} failed to import {package}')
            sys.exit(1)

    return b_module


BIO2BEL_PREFIX = 'bio2bel'


def handle_bio2bel(module_name: str) -> np.ndarray:
    """Load a Bio2BEL repository.

    :param module_name: The name of the bio2bel repository (with no prefix)
    """
    path = install_bio2bel_module(module_name)
    return np.loadtxt(
        fname=path,
        dtype=str,
        comments='@Comment@ Subject Predicate Object',
        delimiter='\t',
    )


def handle_bel_commons(network_id: Union[int, str], host: Optional[str] = None) -> np.ndarray:
    """Load a BEL document from BEL Commons.

    :param network_id: The network identifier in BEL Commons
    :param host: The host for BEL Commons. Defaults to the Fraunhofer SCAI public instance.
    """
    graph = from_web(int(network_id), host=host)
    df = to_pykeen_df(graph)
    return df.to_numpy()



# -*- coding: utf-8 -*-

"""A command line interface for BioKEEN.

Why does this file exist, and why not put this in ``__main__``? You might be tempted to import things from ``__main__``
later, but that will cause problems--the code will get executed twice:
- When you run ``python3 -m biokeen`` python will execute``__main__.py`` as a script. That means there won't be any
  ``biokeen.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``biokeen.__main__`` in ``sys.modules``.
.. seealso:: http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import json
import logging
import os
from typing import List, Optional, TextIO

import click
from click_default_group import DefaultGroup

from bio2bel.constants import get_global_connection
from biokeen.constants import EMOJI, VERSION, biokeen_config
from pykeen.constants import VERSION as PYKEEN_VERSION

connection_option = click.option(
    '-c',
    '--connection',
    default=get_global_connection(),
    show_default=True,
    help='Bio2BEL database connection string',
)


@click.group(cls=DefaultGroup, default_if_no_args=False)
@click.version_option()
def main():  # noqa: D401
    """A command line interface for BioKEEN."""


@main.command()
@connection_option
@click.option('-f', '--config', type=click.File())
@click.option('-r', '--rebuild', is_flag=True)
@click.option('-x', '--no-prompt-bio2bel', is_flag=True)
def start(config: Optional[TextIO], connection: str, rebuild: bool, no_prompt_bio2bel: bool):
    """Start the BioKEEN training pipeline."""
    import pykeen

    if config is not None:
        config = json.load(config)
    else:
        from .prompts import prompt_biokeen_config
        config = prompt_biokeen_config(
            connection=connection,
            rebuild=rebuild,
            do_prompt_bio2bel=(not no_prompt_bio2bel),
        )

    config['pykeen-version'] = PYKEEN_VERSION
    config['biokeen-version'] = VERSION
    pykeen.run(config)


@main.command()
@click.option('-m', '--model-directory', type=click.Path(file_okay=False, dir_okay=True, exists=True), required=True)
@click.option('-d', '--data-directory', type=click.Path(file_okay=False, dir_okay=True, exists=True), required=True)
def predict(model_directory: str, data_directory: str):
    """Use a trained model to make predictions."""
    from pykeen.predict import start_predictions_pipeline
    start_predictions_pipeline(model_directory, data_directory)


@main.group()
def data():
    """Commands for data acquisition."""


@data.command(help=f'Data stored in {biokeen_config.data_directory}')
def ls():
    """List built data."""
    for path in biokeen_config.iterate_source_paths():
        click.echo(path)


@data.command()
@click.confirmation_option()
def clear():
    """Remove all built data."""
    for path in biokeen_config.iterate_source_paths():
        os.remove(path)


@data.command()
@click.argument('names', nargs=-1)
@connection_option
@click.option('-r', '--rebuild', is_flag=True)
@click.option('-v', '--verbose', count=True)
def get(names: List[str], connection: str, rebuild: bool, verbose: bool):
    """Install, populate, and build Bio2BEL repository."""
    if verbose == 1:
        logging.basicConfig(level=logging.INFO)
    elif verbose == 2:
        logging.basicConfig(level=logging.DEBUG)

    from biokeen.content import install_bio2bel_module

    for name in names:
        click.secho(f'{EMOJI} Getting {name}', fg='cyan')
        install_bio2bel_module(name, connection, rebuild)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Prompts for the BioKEEN command line interface."""

import os
import re
from collections import OrderedDict
from typing import Dict, Iterable, Optional

import click

from pykeen.cli.prompt import prompt_config
from pykeen.cli.utils.cli_print_msg_helper import print_section_divider
from pykeen.constants import TRAINING_SET_PATH
from .messages import print_intro, print_welcome_message
from ..constants import ID_TO_DATABASE_MAPPING
from ..content import install_bio2bel_module

__all__ = [
    'prompt_biokeen_config',
]


def prompt_biokeen_config(*, connection: str, rebuild: bool, do_prompt_bio2bel: Optional[bool] = None) -> Dict:
    """Configure experiments."""
    config = OrderedDict()

    # Step 1: Welcome + Intro
    print_welcome_message()
    print_section_divider()
    print_intro()
    print_section_divider()

    # Step 2: Ask for data source
    if do_prompt_bio2bel is None:
        do_prompt_bio2bel = click.confirm('Do you want to use one of the databases provided by BioKEEN?', default=True)
        print_section_divider()

    do_prompt_training = True

    if do_prompt_bio2bel:
        do_prompt_training = False
        config[TRAINING_SET_PATH] = []
        for name in select_bio2bel_repository():
            try:
                path = install_bio2bel_module(name=name, connection=connection, rebuild=rebuild)
            except Exception:
                click.secho(f'failed: {name}', fg='red')
            else:
                if os.path.exists(path):
                    config[TRAINING_SET_PATH].append(f'bio2bel:{name}')
                else:
                    click.secho(f'failed: {name}: {path}', fg='red')

        # TODO replace this with less safe code that assumes everything installs no problemo
        """
        config[TRAINING_SET_PATH] = [
            f'bio2bel:{name}'
            for name in select_bio2bel_repository()
        ]
        """

    print_section_divider()

    return prompt_config(
        config=config,
        show_welcome=False,
        do_prompt_training=do_prompt_training,
    )


def select_bio2bel_repository() -> Iterable[str]:
    """Prompt the user for a Bio2BEL database."""
    click.secho("Current Step: Please select the database(s) you want to train on:", fg='blue')

    number_width = 1 + round(len(ID_TO_DATABASE_MAPPING) / 10)
    for model, model_id in sorted(ID_TO_DATABASE_MAPPING.items()):
        click.echo(f'{model: >{number_width}}: {model_id}')

    while True:
        try:
            user_input = click.prompt('> Please select one or more of the options', value_proc=process_selection)
        except ValueError as e:
            click.secho(str(e), fg='red')
        else:
            return user_input


def process_selection(values: str) -> Iterable[str]:
    for value in re.split(r'\s|,', values):
        value = value.strip()
        if not value:
            continue

        try:
            value = int(value)
        except ValueError:
            if value not in ID_TO_DATABASE_MAPPING.values():
                raise ValueError(f'{value} is an invalid database')
            yield value
        else:
            if value not in ID_TO_DATABASE_MAPPING:
                raise ValueError(f'{value} is an invalid index')
            yield ID_TO_DATABASE_MAPPING[value]



# -*- coding: utf-8 -*-

"""Input and output for BEL conversion."""

import itertools as itt
import json
import logging
from collections import Counter
from typing import Dict, Optional, Tuple

import pandas as pd
from tqdm import tqdm

from pybel import BELGraph
from pybel.dsl import BaseEntity
from .converters import (
    AssociationConverter, CorrelationConverter, DecreasesAmountConverter, DrugIndicationConverter,
    DrugSideEffectConverter, EquivalenceConverter, IncreasesAmountConverter, IsAConverter,
    ListComplexHasComponentConverter, MiRNADecreasesExpressionConverter, MiRNADirectlyDecreasesExpressionConverter,
    NamedComplexHasComponentConverter, PartOfNamedComplexConverter, ProteinPartOfBiologicalProcess,
    RegulatesActivityConverter, RegulatesAmountConverter, SubprocessPartOfBiologicalProcess,
)
from ..constants import EMOJI

__all__ = [
    'to_pykeen_path',
    'to_pykeen_df',
    'get_pykeen_summary',
    'to_pykeen_summary_path',
    'get_triple',
]

logger = logging.getLogger(__name__)


def to_pykeen_path(df: pd.DataFrame, path: str) -> bool:
    """Write the relationships in the BEL graph to a KEEN TSV file.

    If you have a BEL graph, first do:

    >>> from biokeen.convert import to_pykeen_df, to_pykeen_path
    >>> graph = ...  # Something from PyBEL
    >>> df = to_pykeen_df(graph)
    >>> to_pykeen_path(df, 'graph.keen.tsv')
    """
    if len(df.index) == 0:
        return False
    df.to_csv(path, sep='\t', index=None, header=None)
    return True


def get_pykeen_summary(df: pd.DataFrame) -> Dict:
    """Summarize a KEEN dataframe."""
    entity_count = Counter(itt.chain(df[df.columns[0]], df[df.columns[2]]))
    return {
        'namespaces': Counter(
            element.split(':')[0]
            for element in itt.chain(df[df.columns[0]], df[df.columns[2]])
        ),
        'entities': len(entity_count),
        'relations': len(df.index),
    }


def to_pykeen_summary_path(df: pd.DataFrame, path: str, indent=2, **kwargs):
    """Write the summary of a KEEN dataframe to a file."""
    with open(path, 'w') as file:
        json.dump(get_pykeen_summary(df), file, indent=indent, **kwargs)


def to_pykeen_df(graph: BELGraph, use_tqdm: bool = True) -> pd.DataFrame:
    """Get a DataFrame representing the triples."""
    it = graph.edges(keys=True)

    if use_tqdm:
        it = tqdm(it, total=graph.number_of_edges(), desc=f'{EMOJI} preparing TSV')

    triples = (
        get_triple(graph, u, v, key)
        for u, v, key in it
    )

    # clean duplicates and Nones
    triples = list(sorted({triple for triple in triples if triple is not None}))

    return pd.DataFrame(triples, columns=['subject', 'predicate', 'object'])


def get_triple(graph: BELGraph, u: BaseEntity, v: BaseEntity, key: str) -> Optional[Tuple[str, str, str]]:  # noqa: C901
    """Get the triples' strings that should be written to the file."""
    data = graph[u][v][key]

    # order is important
    converters = [
        NamedComplexHasComponentConverter,
        ListComplexHasComponentConverter,
        PartOfNamedComplexConverter,
        SubprocessPartOfBiologicalProcess,
        ProteinPartOfBiologicalProcess,
        RegulatesActivityConverter,
        MiRNADecreasesExpressionConverter,
        MiRNADirectlyDecreasesExpressionConverter,
        IsAConverter,
        EquivalenceConverter,
        CorrelationConverter,
        AssociationConverter,
        DrugIndicationConverter,
        DrugSideEffectConverter,
        RegulatesAmountConverter,
        IncreasesAmountConverter,
        DecreasesAmountConverter,
    ]

    for converter in converters:
        if converter.predicate(u, v, key, data):
            return converter.convert(u, v, key, data)

    logger.warning(f'{EMOJI} unhandled: {graph.edge_to_bel(u, v, data)}')
    return None
